/*
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Local Disk Manager plugin.
 */

#define _GNU_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <mntent.h>
#include <dirent.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <errno.h>

#include <plugin.h>
#include <ldm_dpc.h>
#include "localdskmgr.h"
#include "cache.h"

engine_functions_t *EngFncs = NULL;
BOOLEAN all_acc = FALSE;
BOOLEAN all_dir = FALSE;
list_entry_t *accepts = NULL;
list_entry_t *rejects = NULL;
list_entry_t *dirs = NULL;
char * base_directory = NULL;

BOOLEAN version_2_4 = FALSE;

/**
 * round_down_to_hard_sector
 * @size:	Count of sectors to be rounded down
 * @hs:		Hard-sector-size, in bytes
 *
 * Given a size in vsectors, round it down to be a multiple of the sector size
 * of a given storage object.
 **/
static inline sector_count_t round_down_to_hard_sector(sector_count_t size,
						       u_int32_t hs)
{
	if (hs > EVMS_VSECTOR_SIZE) {
		sector_count_t hs_in_vsectors = ((sector_count_t)hs) >> EVMS_VSECTOR_SIZE_SHIFT;
		return size & ~(hs_in_vsectors - 1);
	} else {
		return size;
	}
}

/**
 * round_up_to_hard_sector
 * @size:	Count of sectors to be rounded up
 * @hs:		Hard-sector-size, in bytes
 *
 * Given a size in vsectors, round it up to be a multiple of the sector size
 * of a given storage object.
 **/
static inline sector_count_t round_up_to_hard_sector(sector_count_t size,
						     u_int32_t hs)
{
	if (hs  > EVMS_VSECTOR_SIZE) {
		sector_count_t hs_in_vsectors = ((sector_count_t)hs) >> EVMS_VSECTOR_SIZE_SHIFT;
		sector_count_t temp_size = size + hs_in_vsectors - 1;
		return temp_size & ~(hs_in_vsectors - 1);
	} else {
		return size;
	}
}

/**
 * is_2_4
 *
 * Determine if we are running on a 2.4 kernel.
 **/
static void is_2_4(void)
{
	int fd;
	char needle[] = "version";
	char haystack[256];
	char *ver;
	int major, minor, patch;

	LOG_PROC_ENTRY();

	fd = open("/proc/version", O_RDONLY);
	if (fd < 0) {
		version_2_4 = TRUE;
		LOG_PROC_EXIT_BOOLEAN(TRUE);
		return;
	}
	read(fd, haystack, 256);
	close(fd);
	ver = strstr(haystack, needle);
	if (!ver) {
		version_2_4 = TRUE;
		LOG_PROC_EXIT_BOOLEAN(TRUE);
		return;
	}
	sscanf(ver, "%*s %d.%d.%d", &major, &minor, &patch);

	LOG_DEBUG("Kernel version is: %d.%d.%d\n", major, minor, patch);
	if (major == 2 && minor == 4) {
		version_2_4 = TRUE;
		LOG_DEBUG("All I/O will be done in at least 1K chunks\n");
	}
	LOG_PROC_EXIT_BOOLEAN(version_2_4);
}

/**
 * where_is_sysfs
 *
 * Is sysfs mounted. If so, return the mount point.
 **/
static BOOLEAN where_is_sysfs(char ** mount_name)
{
	BOOLEAN result = FALSE;
	FILE * mount_records;
	struct mntent * mount_entry;

	LOG_PROC_ENTRY();

	if (mount_name != NULL) {
		*mount_name = NULL;
	}

	mount_records = setmntent(MOUNTED, "r");
	if (mount_records == NULL) {
		mount_records = setmntent("/proc/mounts", "r");
	}
	if (mount_records == NULL) {
		LOG_ERROR("Could not get list of mounted devices.\n");
		LOG_PROC_EXIT_BOOLEAN(result);
		return result;
	}
	while (!result && (mount_entry = getmntent(mount_records)) != NULL) {
		if (strcmp(mount_entry->mnt_type, "sysfs") == 0) {
			result = TRUE;
			if (mount_name != NULL) {
				*mount_name = strdup( mount_entry->mnt_dir );
			}
		}
	}

	endmntent(mount_records);

	LOG_PROC_EXIT_BOOLEAN(result);
	return result;
}

/** 
 * allocate_and_insert
 *
 * Allocate a list element, copy the name, and insert and the front of the
 * specified list.
 **/
static int allocate_and_insert(const char * name, list_entry_t ** head)
{
	list_entry_t * le = NULL;

	LOG_PROC_ENTRY();

	le = EngFncs->engine_alloc(sizeof(list_entry_t));
	if (!le) {
		LOG_PROC_EXIT_INT(ENOMEM);
		return ENOMEM;
	}

	LOG_DEBUG("Adding %s to list.\n", name);

	strncpy(le->entry, name, EVMS_NAME_SIZE);
	le->next = *head;
	le->prev = NULL;
	if (le->next) {
		le->next->prev = le;
	}
	*head = le;

	LOG_PROC_EXIT_INT(0);
	return 0;
}

/**
 * rejects_config
 *
 * Set up the default entries to reject, and see if the user specified any
 * additional entries to reject.
 **/
static int rejects_config(ver_t ver)
{
	int rc = EINVAL, i, count=0;
	const char * const * rej = NULL;

	LOG_PROC_ENTRY();

	/* Version independent. */
	allocate_and_insert("..", &rejects);
	allocate_and_insert(".", &rejects);

	if (ver == LEGACY) {	
		rc = EngFncs->get_config_string_array("legacy_devices.exclude",
						      &count, &rej);
	} else if (ver == SYSFS) {
		allocate_and_insert("fd*", &rejects);
		allocate_and_insert("md*", &rejects);
		allocate_and_insert("dm*", &rejects);
		allocate_and_insert("evms*", &rejects);
		rc = EngFncs->get_config_string_array("sysfs_devices.exclude",
						      &count, &rej);
	}

	if (rc) {
		/* User has specified no additional rejects. */
		LOG_DEBUG("config exclude returned: %d, version: %d\n", rc, ver);
		LOG_PROC_EXIT_INT(0);
		return 0;
	}

	for (i = 0; i < count; i++) {
		if (strcmp(rej[i], "*") == 0) {
			LOG_ERROR("Config file error, cannot reject all\n");
			LOG_PROC_EXIT_INT(EINVAL);
			return EINVAL;
		}
		allocate_and_insert(rej[i], &rejects);
	}

	LOG_PROC_EXIT_INT(0);
	return 0;
}

/**
 * accepts_config
 *
 * Set up the device entries to accept.
 **/
static int accepts_config(ver_t ver)
{
	int rc, i, count = 0;
	const char * const * acc = NULL;

	LOG_PROC_ENTRY();

	if (ver == LEGACY) {
		rc = EngFncs->get_config_string_array("legacy_devices.include",
						      &count, &acc);
		if (rc || count == 0) {
			LOG_DEBUG("config include returned: %d\n", rc);
			allocate_and_insert("hd*", &accepts);
			allocate_and_insert("sd*", &accepts);
			allocate_and_insert("dasd*", &accepts);
			LOG_PROC_EXIT_INT(0);
			return 0;
		}
	} else if (ver == SYSFS) {
		rc = EngFncs->get_config_string_array("sysfs_devices.include",
						      &count, &acc);
		if (rc || count == 0) {
			/* Accept all devices. */
			LOG_DEBUG("config include returned: %d\n", rc);
			all_acc = TRUE;
			LOG_PROC_EXIT_INT(0);
			return 0;
		}
	}

	for (i = 0; i < count; i++) {
		if (strcmp(acc[i], "*") == 0) {
			/* Accept all devices. */
			all_acc = TRUE;
			LOG_PROC_EXIT_INT(0);
			return 0;
		}
		allocate_and_insert(acc[i], &accepts);
	}

	LOG_PROC_EXIT_INT(0);
	return 0;
}

/**
 * check_rejects
 *
 * Check if the specified device has been marked as rejected.
 **/
static int check_rejects(const struct dirent * d)
{
	list_entry_t * curr = rejects;

	while (curr) {
		char * p_star = strchr(curr->entry, '*');

		if (p_star == NULL) {
			if (strcmp(d->d_name, curr->entry) == 0) {
				/* Rejected. */
				LOG_DEBUG("%s rejected.\n", d->d_name);
				return 0;
			}
		} else {
			if (strncmp(d->d_name, curr->entry, p_star - curr->entry) == 0) {
				/* Rejected. */
				LOG_DEBUG("%s rejected.\n", d->d_name);
				return 0;
			}
		}
		curr = curr->next;
	}

	/* Accepted. */
	LOG_DEBUG("%s accepted.\n", d->d_name);
	return 1;
}

/**
 * check_accepts
 *
 * Check if the specified device has been marked as accepted.
 **/
static int check_accepts(const struct dirent * d)
{
	list_entry_t * curr = accepts;

	if (all_acc) {
		return check_rejects(d);
	}

	while (curr) {
		char * p_star = strchr(curr->entry, '*');

		if (p_star == NULL) {
			if (strcmp(d->d_name, curr->entry) == 0) {
				return check_rejects(d);
			}
		} else {
			if (strncmp(d->d_name, curr->entry, p_star - curr->entry) == 0) {
				return check_rejects(d);
			}
		}
		curr = curr->next;
	}

	/* Rejected. */
	LOG_DEBUG("%s rejected.\n", d->d_name);
	return 0;
}

/**
 * close_dev
 **/
static int close_dev(void * object,
		     TAG object_tag,
		     void * object_handle,
		     void * parameters)
{
	storage_object_t * obj = object;
	int * fd = obj->private_data;

	LOG_PROC_ENTRY();

	/* Safety check. */
	if (object_tag == DISK_TAG) {
		if (fd) {
			EngFncs->close_object(obj, *fd);
			obj->private_data = NULL;
		}
	}

	LOG_PROC_EXIT_INT(DLIST_SUCCESS);
	return DLIST_SUCCESS;
}


/********** Legacy Version **********/


/**
 * dirs_config
 **/
static int dirs_config(void)
{
	int rc, i, count;
	const char * const * dir = NULL;

	LOG_PROC_ENTRY();

	rc = EngFncs->get_config_string_array("legacy_devices.directories",
					      &count, &dir);
	if (rc) {
		LOG_PROC_EXIT_INT(0);
		return 0;
	}

	for (i = 0; i < count; i++) {
		if (strcmp(dir[i], "*") == 0) {
			/* Accept all. */
			all_dir = TRUE;
			LOG_PROC_EXIT_INT(0);
			return 0;
		}
		allocate_and_insert(dir[i], &dirs);
	}

	LOG_PROC_EXIT_INT(0);
	return 0;
}

/**
 * check_dirs
 *
 * Check the specified directory to see if it is accepted.
 **/
static int check_dirs(const struct dirent * d)
{
	list_entry_t * curr = dirs;

	if (strcmp(d->d_name, "..") == 0     ||
	    strcmp(d->d_name, "." ) == 0     ||
	    strcmp(d->d_name, "mapper") == 0 ||
	    strcmp(d->d_name, "evms") == 0) {
		return 0;
	}

	if (all_dir) {
		/* Accept all. */
		return 1;
	}

	while (curr) {
		if (strncmp(d->d_name, curr->entry, strlen(curr->entry)) == 0) {
			/* Accepted. */
			return 1;
		}
		curr = curr->next;
	}

	/* Rejected. */
	return 0;
}

/**
 * check_subdirs
 * 
 * Accept all sub-directories except "." and "..".
 **/
static int check_subdirs(const struct dirent * d)
{
	if (strcmp(d->d_name, "..") == 0 ||
	    strcmp(d->d_name, "." ) == 0) {
		return 0;
	}

	return 1;
}

/**
 * block_select
 **/
static int block_select(const char * dir, dlist_t disklist)
{
	storage_object_t working_object, *ld;
	struct dirent ** namelist = NULL;
	struct hd_geometry geo;
	struct stat st;
	void * handle;
	char name[256];
	char * subdir_name;
	u_int32_t hardsector_size;
	int rc, n = 0, dev_handle = 0;

	LOG_PROC_ENTRY();

	subdir_name = strstr(dir, base_directory);
	if (subdir_name) {
		subdir_name += strlen(base_directory);
	} else {
		subdir_name = (char *)dir;
	}

	n = scandir(dir, &namelist, check_accepts, NULL);
	if (n < 0) {
		rc = errno;
		LOG_ERROR("Encountered error(%d) in scandir of %s during "
			  "discover\n", rc, dir);
		LOG_PROC_EXIT_INT(rc);
		return rc;
	}

	/* Initialize working object.
	 * Will need to initialize other fields for clustering.
	 */
	memset(&working_object, 0, sizeof(working_object));
	working_object.object_type = DISK;
	working_object.data_type = DATA_TYPE;
	working_object.plugin = plugin_legacy;
	
	while (n--) {
		LOG_DEBUG("Working on object: %s\n", namelist[n]->d_name);
		strncpy(working_object.name, subdir_name, EVMS_NAME_SIZE);
		strncat(working_object.name, namelist[n]->d_name,
			EVMS_NAME_SIZE-strlen(working_object.name));
		free(namelist[n]);
	
		/* name = dir + namelist[n]->d_name */
		strncpy(name, base_directory, 256);
		strncat(name, working_object.name, 256-strlen(name));

		/* Exclude things which fail stat or
		 * which are not block devices.
		 */
		if (stat(name, &st) < 0 || ! S_ISBLK(st.st_mode)) {
			continue;
		}
	
		/* Exclude floppys(2 & 47), md's(9), and lvm1's(58). */
		working_object.dev_major = major(st.st_rdev);
		if (working_object.dev_major ==  2 ||
		    working_object.dev_major ==  9 ||
		    working_object.dev_major == 47 ||
		    working_object.dev_major == 58 ) {
			continue;
		}

		/* Open test. */
		working_object.dev_minor = minor(st.st_rdev);
		LOG_DEBUG("Major: %u Minor: %u\n",
			  working_object.dev_major, working_object.dev_minor);

		dev_handle = EngFncs->open_object(&working_object,
						  O_DIRECT|O_RDWR|O_SYNC);
		if (dev_handle < 0) {
			LOG_DEBUG("Unable to open, cannot create a disk object.\n");
			continue;
		}

		/* Ioctl to get size. (returns bytes) */
		rc = EngFncs->ioctl_object(&working_object, dev_handle,
					   BLKGETSIZE64, &working_object.size);
		if (rc) {
			LOG_DEBUG("Error getting size, cannot create disk object.\n");
			goto out;
		}

		working_object.size >>= EVMS_VSECTOR_SIZE_SHIFT;
		if (working_object.size <= 0) {
			LOG_DEBUG("Size of zero cannot be a real disk.\n");
			goto out;
		}

		LOG_DEBUG("Size: %"PRIu64"\n", working_object.size);
	
		/* Ioctl to get geo check for a start at zero. */
		rc = EngFncs->ioctl_object(&working_object, dev_handle,
					   HDIO_GETGEO, &geo);
		if (rc) {
			LOG_DEBUG("Error getting geometry, cannot create disk object.\n");
			goto out;
		}
		if (geo.start != 0) {
			LOG_DEBUG("Geometry says start not at zero; therefore, not considered a disk\n");
			goto out;
		}

		/* Ioctl to get hardsector size. */
		rc = EngFncs->ioctl_object(&working_object, dev_handle,
					   BLKSSZGET, &hardsector_size);
		if (rc) {
			LOG_DEBUG("Error getting hardsector size, cannot create disk object.\n");
			goto out;
		}

#if 0
		/* This section not need did a stat above */

		/* fstat to get Blocksize */
		rc = fstat(dev_handle, &st);
		if (rc) {
			LOG_DEBUG("Error with fstat, cannot create a disk object.\n");
			goto out;
		}
#endif

		/* Get a logical_disk struct */
		rc = EngFncs->allocate_logical_disk(working_object.name, &ld);
		if (rc) {
			LOG_DEBUG("Error code %d from Engine's allocate_logical_disk, cannot create a disk object.\n", rc);
			goto out;
		}

		ld->private_data = EngFncs->engine_alloc(sizeof(int));
		if (!ld->private_data) {
			LOG_DEBUG("Cannot allocate private data for disk %s.\n", ld->name);
			EngFncs->free_logical_disk(ld);
			goto out;
		}

		*(int *)ld->private_data	= dev_handle;

		ld->plugin			= plugin_legacy;
		ld->data_type			= DATA_TYPE;
		ld->flags			|= SOFLAG_ACTIVE;
		ld->dev_major			= working_object.dev_major;
		ld->dev_minor			= working_object.dev_minor;
		ld->size			= working_object.size;
		ld->geometry.cylinders		= geo.cylinders;
		ld->geometry.heads		= geo.heads;
		ld->geometry.sectors_per_track	= geo.sectors;
		ld->geometry.bytes_per_sector	= hardsector_size;
		ld->geometry.block_size		= st.st_blksize;

		/* LBA of 1st sector above boot cylinder limit for this drive  */
		if (ld->geometry.cylinders < 1024) {
			/* If drive is too small,
			 * then the limit is the size of the drive.
			 */
			ld->geometry.boot_cylinder_limit = ld->size;
		} else {
			/* else the limit is calculated. */
			ld->geometry.boot_cylinder_limit =
				ld->geometry.heads * ld->geometry.sectors_per_track * 1023;
		}
	
		/* Insert the logical disk into our discovered disk list */
		rc = InsertObject(disklist, ld, DISK_TAG, NULL, AppendToList, TRUE, &handle);
		if (rc) {
			EngFncs->free_logical_disk(ld);
			LOG_DEBUG("Error code %d when inserting new logical disk into the disk list, cannot create disk object.\n", rc);
			goto out;
		}

		/* Successful */
		LOG_DETAILS("New Logical Device:\n");
		LOG_DETAILS("               name: %s\n", ld->name);
		LOG_DETAILS("               size: %"PRIu64"\n", ld->size);
		LOG_DETAILS("     geometry:\n");
		LOG_DETAILS("          cylinders: %"PRIu64"\n", ld->geometry.cylinders);
		LOG_DETAILS("              heads: %d\n", ld->geometry.heads);
		LOG_DETAILS("            sectors: %d\n", ld->geometry.sectors_per_track);
		LOG_DETAILS("        sector size: %d (bytes)\n", ld->geometry.bytes_per_sector);

		/* Skip error condition cleanup. */
		continue;

	/* Error condition cleanup. */
	out:
		LOG_DEBUG("closing: %d\n", dev_handle);
		EngFncs->close_object(&working_object, dev_handle);
		dev_handle = 0;
	}

	if (namelist) {
		free(namelist);
	}

	LOG_PROC_EXIT_INT(0);
	return 0;
}

/**
 * dir_select
 **/
static int dir_select(const char * dir, dlist_t disklist, int first_pass)
{
	char name[256];
	struct stat st;
	int rc, n = 0;
	struct dirent ** namelist = NULL;

	LOG_PROC_ENTRY();

	if (first_pass) {
		n = scandir(dir, &namelist, check_dirs, NULL);
	} else {
		n = scandir(dir, &namelist, check_subdirs, NULL);
	}

	if (n < 0) {
		rc = errno;
		LOG_ERROR("Encounter error(%d) in scandir of %s during discover\n",
			  rc, dir);
		LOG_PROC_EXIT_INT(rc);
		return rc;
	}

	block_select(dir, disklist);

	while (n--) {
		/* Directory check. */
		strncpy(name, dir, 256);
		strncat(name, namelist[n]->d_name, 256-strlen(name));
		free(namelist[n]);

		if (stat(name, &st) || !S_ISDIR(st.st_mode)) {
			continue;
		}

		/* Recurse */
		strncat(name, "/", 256-strlen(name));
		dir_select(name, disklist, 0);
	}

	free(namelist);
	LOG_PROC_EXIT_INT(0);
	return 0;
}


/********** SysFS Version **********/ 


/**
 * discovery_select
 **/
static int discovery_select(const char * dir, dlist_t disk_list)
{
	struct dirent ** namelist = NULL;
	storage_object_t working_object, *ld;
	int dev_handle = 0, n = 0, rc = 0;
	FILE *sys;
	char buf[256], name[EVMS_NAME_SIZE+1], file[EVMS_NAME_SIZE+1];
	void * handle;
	struct hd_geometry geo;
	struct stat st;
	u_int32_t hardsector_size;
	dev_t dev;

	LOG_PROC_ENTRY();

	n = scandir(dir, &namelist, check_accepts, NULL);
	if (n < 0) {
		rc = errno;
		LOG_ERROR("Encounter error(%d) in scandir of %s during discover\n",
			  rc, dir );
		LOG_PROC_EXIT_INT(rc);
		return rc;
	}

	/* Initialize working object.
	 * Will need to initialize other fields for clustering.
	 */
	memset(&working_object, 0, sizeof(working_object));
	working_object.object_type = DISK;
	working_object.data_type = DATA_TYPE;
	working_object.plugin = plugin_sysfs;

	/* Get logical disks and insert into our dlist_t. */
	while (n--) {
		LOG_DEBUG("Working on object: %s\n", namelist[n]->d_name);
		strncpy(working_object.name, namelist[n]->d_name, EVMS_NAME_SIZE);
		/* Cleanup is done don't use namelist element anymore. */
		free(namelist[n]);

		/* Build name once per loop to equal where to read the sysfs files. */
		strncpy(name, dir, EVMS_NAME_SIZE);
		strncat(name, working_object.name, EVMS_NAME_SIZE-strlen(name));

		/* Exclude things which fail stat or which are not directories. */
		if (stat(name, &st) < 0 || ! S_ISDIR(st.st_mode)) {
			continue;	
		}

		/* Get size and check that it's greater than zero. */
		strncpy(file, name, EVMS_NAME_SIZE);
		strncat(file, "/size", EVMS_NAME_SIZE-strlen(file));
		sys = fopen(file, "r");
		fscanf(sys, "%s", buf);
		fclose(sys);

		working_object.size = strtoull(buf, NULL, 10); /* sectors */
		LOG_DEBUG("Size: %"PRIu64"\n", working_object.size);

		if (working_object.size <= 0) {
			LOG_DEBUG("Size of zero cannot be a real disk.\n");
			continue;
		}
	
		/* Get the device's major and minor for creating the dev node. */
		strncpy(file, name, EVMS_NAME_SIZE);
		strncat(file, "/dev", EVMS_NAME_SIZE-strlen(file));
		sys = fopen(file, "r");
		fscanf(sys, "%s", buf);
		fclose(sys);
	
		dev = strtoul(buf, NULL, 16);
		working_object.dev_major = major(dev);
		working_object.dev_minor = minor(dev);
		LOG_DEBUG("Major: %u Minor: %u\n",
			  working_object.dev_major, working_object.dev_minor);

		/* Use Engine to Create and Open. */
		dev_handle = EngFncs->open_object(&working_object,
						  O_DIRECT|O_RDWR|O_SYNC);
		if (dev_handle < 0) {
			LOG_DEBUG("Unable to open, cannot create a disk object.\n");
			dev_handle = 0;
			continue;
		}

		/* fstat to get blocksize */
		rc = fstat(dev_handle, &st);
		if (rc) {
			LOG_DEBUG("Error with fstat, cannot create a disk object.\n");
			goto out;
		}

		/* Ioctl to get geometry */
		rc = EngFncs->ioctl_object(&working_object, dev_handle,
					   HDIO_GETGEO, &geo);
		if (rc) {
			LOG_DEBUG("Error getting geometry, cannot create a disk object.\n");
			goto out;
		}
		if (geo.start != 0) {
			LOG_DEBUG("Geometry says start not at zero; therefore, not considered a disk\n");
			goto out;
		}
	
		/* Ioctl to get hardsector size */
		rc = EngFncs->ioctl_object(&working_object, dev_handle,
					   BLKSSZGET, &hardsector_size);
		if (rc) {
			LOG_DEBUG("Error getting hardsector size, cannot create a disk object.\n");
			goto out;
		}

		/* Passed all checks Get a logical_disk struct */
		rc = EngFncs->allocate_logical_disk(working_object.name, &ld);
		if (rc) {
			LOG_DEBUG("Error code %d from Engine's allocate_logical_disk, cannot create a disk object.\n", rc);
			goto out;
		}

		ld->private_data = EngFncs->engine_alloc(sizeof(int));
		if (!ld->private_data) {
			LOG_DEBUG("Cannot allocate private data for disk %s.\n", ld->name);
			EngFncs->free_logical_disk(ld);
			goto out;
		}

		*(int *)ld->private_data	= dev_handle;

		/* Initialize the logical disk structure */
		ld->plugin			= plugin_sysfs;
		ld->data_type			= DATA_TYPE;
		ld->flags			|= SOFLAG_ACTIVE;
		ld->size			= working_object.size;
		ld->dev_major			= working_object.dev_major;
		ld->dev_minor			= working_object.dev_minor;
		ld->geometry.cylinders		= geo.cylinders;
		ld->geometry.heads		= geo.heads;
		ld->geometry.sectors_per_track	= geo.sectors;
		ld->geometry.bytes_per_sector	= hardsector_size;
		ld->geometry.block_size		= st.st_blksize;

		/* LBA of 1st sector above boot cylinder limit for this drive */
		if (ld->geometry.cylinders < 1024) {
			/* If drive is too small
			 * then the limit is the size of the drive.
			 */
			ld->geometry.boot_cylinder_limit = ld->size;
		} else {
			/* else the limit is calculated */
			ld->geometry.boot_cylinder_limit =
				ld->geometry.heads * ld->geometry.sectors_per_track * 1023;
		}

		/* Insert the logical disk into discovered list */
		rc = InsertObject (disk_list, ld, DISK_TAG, NULL, AppendToList, TRUE, &handle);
		if (rc) {
			EngFncs->free_logical_disk(ld);
			LOG_DEBUG("Error code %d when inserting new logical disk into the disk list, cannot create disk object\n", rc);
			goto out;
		}
	
		/* Successful */
		LOG_DETAILS("New Logical Device:\n");
		LOG_DETAILS("               name: %s\n", ld->name);
		LOG_DETAILS("               size: %"PRIu64"\n", ld->size);
		LOG_DETAILS("     geometry:\n");
		LOG_DETAILS("          cylinders: %"PRIu64"\n", ld->geometry.cylinders);
		LOG_DETAILS("              heads: %d\n", ld->geometry.heads);
		LOG_DETAILS("            sectors: %d\n", ld->geometry.sectors_per_track);
		LOG_DETAILS("        sector size: %d (bytes)\n", ld->geometry.bytes_per_sector);

		/* Skip error condition cleanup. */
		continue;

	/* Error condition cleanup */
	out:
		LOG_DEBUG("closing: %d\n", dev_handle);
		EngFncs->close_object(&working_object, dev_handle);
		dev_handle = 0;
	}

	if (namelist) {
		free(namelist);
	}
	LOG_PROC_EXIT_INT(0);
	return 0;
}

/**
 * LD_setup_legacy
 **/
static int LD_setup_legacy(engine_functions_t * engine_function_table)
{
	int rc;

	if (LD_PluginRecord) {
		return E_NOLOAD;
	}

	/* save info we get from the engine */
	EngFncs = engine_function_table;
	LD_PluginRecord = plugin_legacy;

	LOG_PROC_ENTRY();

	if (where_is_sysfs(NULL)) {
		/* "sysfs" was mounted. Don't want this plugin */
		rc = E_NOLOAD;
		goto out;
	}

	/* For I/O path need to know if version is 2.4 */
	is_2_4();

	/* sysfs isn't mounted have to use this plugin */
	LOG_WARNING("sysfs isn't mounted. Using the legacy plugin which is less effiecient\n");

	rc = accepts_config(LEGACY);
	if (rc) {
		goto out;
	}

	rc = rejects_config(LEGACY);
	if (rc) {
		goto out;
	}

	rc = dirs_config();
	if (rc) {
		goto out;
	}

	LOG_PROC_EXIT_INT(0);
	return 0;

out:
	LOG_PROC_EXIT_INT(rc);
	LD_PluginRecord = NULL;
	return rc;
}

/**
 * LD_discover_legacy
 *
 * Note: Device Managers ignore the input_list parameter, and only generate
 *       output.
 **/
static int LD_discover_legacy(dlist_t input_list,
			      dlist_t output_list,
			      BOOLEAN final_call)
{
	int len, rc;
	uint count;
	char * tmp = NULL;
	char def[]= "/dev/";
	const char * dir = NULL;

	LOG_PROC_ENTRY();

	/* Get the directory name to scan for disks. */
	rc = EngFncs->get_config_string("legacy_devices.scan", &dir);
	if (rc) {
		base_directory = def;
	} else {
		/* Make sure the user-specified directory ends with a '/'. */
		len = strlen(dir);
		if (dir[len-1] != '/') {
			tmp = malloc(len + 2);
			if (!tmp) {
				dir = def;
			} else {
				strcpy(tmp, dir);
				strcat(tmp, "/");
				dir = tmp;
			}
		}
		base_directory = (char *)dir;
	}

	/* Discover disks in the specified directory. */
	rc = dir_select(base_directory, output_list, TRUE);
	if (tmp) {
		free(tmp);
	}

	GetListSize(output_list, &count);
	LOG_DEBUG("Exporting %d logical disks\n", count);
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * LD_setup_sysfs
 **/
static int LD_setup_sysfs(engine_functions_t * engine_function_table)
{
	int rc = E_NOLOAD;

	if (LD_PluginRecord) {
		return rc;
	}

	/* save info we get from the engine */
	EngFncs = engine_function_table;
	LD_PluginRecord = plugin_sysfs;

	LOG_PROC_ENTRY();

	if (!where_is_sysfs(NULL)) {
		/* "sysfs" was not mounted. Can't use this plugin */
		goto out;
	}

	/* For I/O path need to know if version is 2.4. */
	is_2_4();

	/* sysfs is mounted using the sysfs (more efficient) plugin */
	rc = accepts_config(SYSFS);
	if (rc) {
		goto out;
	}

	rc = rejects_config(SYSFS);
	if (rc) {
		goto out;
	}

	LOG_PROC_EXIT_INT(0);
	return 0;

out:
	LOG_PROC_EXIT_INT(rc);
	LD_PluginRecord = NULL;
	return rc;
}

/**
 * LD_discover_sysfs
 **/
static int LD_discover_sysfs(dlist_t input_list,
			     dlist_t output_list,
			     BOOLEAN final_call)
{
	int rc;
	char *mntpt, dir[EVMS_NAME_SIZE+1];
	uint count;

	LOG_PROC_ENTRY();

	/* Note:  Device Managers ignore the input_list parameter.
	 * Device Managers only generate output.
	 */

	if (!where_is_sysfs(&mntpt)) {
		LOG_ERROR("sysfs not mounted cannot discover with this plugin\n");
		LOG_PROC_EXIT_INT(EINVAL);
		return EINVAL;
	}

	LOG_DEBUG("scanning %s\n", mntpt);
	strncpy(dir, mntpt, EVMS_NAME_SIZE);
	strncat(dir, "/block/", EVMS_NAME_SIZE-strlen(dir));
	free(mntpt);

	rc = discovery_select(dir, output_list);

	GetListSize(output_list, &count);
	LOG_DEBUG("Exporting %d Objects\n", count);
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * LD_cleanup
 *
 * Find any disks & close the device that was opended during discovery.
 * Used by both Legacy and Sysfs versions.
 **/
static void LD_cleanup(void)
{
	int rc;
	dlist_t disk_list;

	LOG_PROC_ENTRY();

	/* Get a list of disks that are managed by this plug-in. */
	rc = EngFncs->get_object_list(DISK,0, LD_PluginRecord, NULL, 0, &disk_list);
	if (!rc) {
		/* Close and dev handles that might be open. */
		ForEachItem(disk_list, close_dev, NULL, TRUE);
		DestroyList(&disk_list, FALSE);
	}

	destroy_cache();

	LOG_PROC_EXIT_VOID();
}

/**
 * LD_read
 *
 * Used by both Legacy and Sysfs versions.
 **/
static int LD_read(storage_object_t * ld,
		   lsn_t offset,
		   sector_count_t count,
		   void * buffer)
{
	void * local_buffer = NULL, * local_buffer_aligned;
	lsn_t local_offset;
	sector_count_t local_count;
	u_int32_t hss, local_buffer_bytes;
	u_int32_t offset_diff, offset_diff_bytes;
	u_int32_t alignment_base, alignment_diff;
	int * fd_p = ld->private_data;
	int fd, rc = EINVAL;

	LOG_PROC_ENTRY();

        if (offset + count > ld->size) {
		LOG_ERROR("Read request at sector %"PRIu64" for %"PRIu64
			  " sectors goes off end of disk.\n", offset, count);
		LOG_ERROR("  Disk: %s   Size: %"PRIu64"\n", ld->name, ld->size);
		goto out;
	}

	rc = read_from_cache(ld, offset, count, buffer);
	if (!rc) {
		/* Found in the cache. */
		goto out;
	}

	/* Not in the cache. Have to go to disk. */
	LOG_DEBUG("Read %"PRIu64" sectors from disk %s at sector %"PRIu64".\n",
		  count, ld->name, offset);

	if (!fd_p) {
		/* Private data must have gotten lost.
		 * Just allocate a new one.
		 */
		fd_p = EngFncs->engine_alloc(sizeof(int));
		if (!fd_p) {
			rc = ENOMEM;
			goto out;
		}
		ld->private_data = fd_p;
	}

	fd = *fd_p;
	if (fd <= 0) {
		/* Need to open the disk. */
		fd = EngFncs->open_object(ld, O_DIRECT|O_RDONLY|O_SYNC);
		if (fd > 0) {
			*fd_p = fd;
		}
	}

	if (fd < 0) {
		/* Couldn't open the disk. */
		rc = EIO;
		goto out;
	}

	/* Grab the hard-sector-size. */
	hss = ld->geometry.bytes_per_sector;
	if (version_2_4 && hss < 1024) {
		/* For O_DIRECT */
		hss = 1024;
	}

	/* Round down starting offset to hard-sector-size. */
	local_offset = round_down_to_hard_sector(offset, hss);

	/* Difference between real offset and local offset. */
	offset_diff = offset - local_offset;
	offset_diff_bytes = offset_diff << EVMS_VSECTOR_SIZE_SHIFT;

	/* Round up total count of sectors to hard-sector-size. */
	local_count = round_up_to_hard_sector(count + offset_diff, hss);

	/* Allocate the buffer that will actually perform the I/O. This needs
	 * to allocate enough extra to compensate for the alignment that will
	 * occur next.
	 */
	local_buffer_bytes = local_count << EVMS_VSECTOR_SIZE_SHIFT;
	local_buffer = EngFncs->engine_alloc(local_buffer_bytes + hss - 1);
	if (!local_buffer) {
		rc = ENOMEM;
		goto out;
	}

	/* Align the buffer on hard-sector-size. */
	alignment_base = (unsigned long)local_buffer & (hss - 1);
	alignment_diff = (alignment_base + hss - 1) & ~(hss - 1);
	alignment_diff -= alignment_base;
	local_buffer_aligned = local_buffer + alignment_diff;

	/* Sanity check */
	if (((unsigned long)local_buffer_aligned % hss) != 0) {
		LOG_DEBUG("BUFFER NOT ALIGNED!!: buf: %p, hss: %u\n",
			  local_buffer_aligned, hss);
	}

	LOG_DEBUG("read(Fd: %d, Buf: %p, Size: %u, Offset: %"PRIu64"\n",
		  fd, local_buffer_aligned, local_buffer_bytes,
		  local_offset << EVMS_VSECTOR_SIZE_SHIFT);

	rc = EngFncs->read_object(ld, fd, local_buffer_aligned, local_buffer_bytes,
				  local_offset << EVMS_VSECTOR_SIZE_SHIFT);
	if (rc < 0) {
		rc = EIO;
		goto out;
	}

	/* Copy the data back to the caller's buffer. */
	memcpy(buffer, local_buffer_aligned+offset_diff_bytes,
	       count << EVMS_VSECTOR_SIZE_SHIFT);
	write_to_cache(ld, offset, count, buffer);
	rc = 0;

out:
	if (local_buffer) {
		EngFncs->engine_free(local_buffer);
	}
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * LD_write
 *
 * Used by both Legacy and Sysfs versions.
 **/
static int LD_write(storage_object_t * ld,
		    lsn_t offset,
		    sector_count_t count,
		    void * buffer)
{
	void * local_buffer = NULL, * local_buffer_aligned;
	lsn_t local_offset;
	sector_count_t local_count;
	u_int32_t hss, local_buffer_bytes;
	u_int32_t offset_diff, offset_diff_bytes;
	u_int32_t alignment_base, alignment_diff;
	int * fd_p = ld->private_data;
	int fd, rc = EINVAL;

	LOG_PROC_ENTRY();

	if (offset + count > ld->size) {
		LOG_ERROR("Write request at sector %"PRIu64" for %"PRIu64
			  " sectors goes off end of disk.\n", offset, count);
		LOG_ERROR("  Disk: %s   Size: %"PRIu64"\n", ld->name, ld->size);
		goto out;
	}

	LOG_DEBUG("Write %"PRIu64" sectors to disk %s at sector %"PRIu64".\n",
		  count, ld->name, offset);

	if (!fd_p) {
		/* Private data must have gotten lost.
		 * Just allocate a new one.
		 */
		fd_p = EngFncs->engine_alloc(sizeof(int));
		if (!fd_p) {
			rc = ENOMEM;
			goto out;
		}
		ld->private_data = fd_p;
	}

	fd = *fd_p;
	if (fd <= 0) {
		/* Need to open the disk. */
		fd = EngFncs->open_object(ld, O_DIRECT|O_RDWR|O_SYNC);
		if (fd > 0) {
			*fd_p = fd;
		}
	}

	if (fd < 0) {
		/* Couldn't open the disk. */
		rc = EIO;
		goto out;
	}

	/* Grab the hard-sector-size. */
	hss = ld->geometry.bytes_per_sector;
	if (version_2_4 && hss < 1024) {
		/* For O_DIRECT */
		hss = 1024;
	}

	/* Round down starting offset to hard-sector-size. */
	local_offset = round_down_to_hard_sector(offset, hss);
	
	/* Difference between real offset and local offset. */
	offset_diff = offset - local_offset;
	offset_diff_bytes = offset_diff << EVMS_VSECTOR_SIZE_SHIFT;
	
	/* Round up total count of sectors to hard-sector-size. */
	local_count = round_up_to_hard_sector(count + offset_diff, hss);
	
	/* Allocate the buffer that will actually perform the I/O. This needs
	 * to allocate enough extra to compensate for the alignment that will
	 * occur next.
	 */
	local_buffer_bytes = local_count << EVMS_VSECTOR_SIZE_SHIFT;
	local_buffer = EngFncs->engine_alloc(local_buffer_bytes + hss - 1);
	if (!local_buffer) {
		rc = ENOMEM;
		goto out;
	}
	
	/* Align the buffer on hard-sector-size. */
	alignment_base = (unsigned long)local_buffer & (hss - 1);
	alignment_diff = (alignment_base + hss - 1) & ~(hss - 1);
	alignment_diff -= alignment_base;
	local_buffer_aligned = local_buffer + alignment_diff;

	/* Sanity check */
	if (((unsigned long)local_buffer_aligned % hss) != 0) {
		LOG_DEBUG("BUFFER NOT ALIGNED!!: buf: %p, hss: %u\n",
			  local_buffer_aligned, hss);
	}

	if (local_count != count) {
		/* Padding for alignment read first */
		rc = EngFncs->read_object(ld, fd, local_buffer_aligned,
					  local_buffer_bytes,
					  local_offset << EVMS_VSECTOR_SIZE_SHIFT);
		if (rc < 0) {
			rc = EIO;
			goto out;
		}
	}
	
	/* Put user data at the right place in the buffer */
	memcpy(local_buffer_aligned+offset_diff_bytes, buffer,
	       count << EVMS_VSECTOR_SIZE_SHIFT);
	
	rc = EngFncs->write_object(ld, fd, local_buffer_aligned,
				   local_buffer_bytes,
				   local_offset << EVMS_VSECTOR_SIZE_SHIFT);
	if (rc < 0) {
		rc = EIO;
		goto out;
	}

	/*
	 * The cache is too simple to do real caching.  It's really only a read
	 * cache.  A write, which should not happen during discovery, means the
	 * contents of the cache may not be up to date.  Purge the cache and
	 * start caching all over again.
	 */
	purge_cache();
	rc = 0;

out:
	if (local_buffer) {
		EngFncs->engine_free(local_buffer);
	}
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * LD_add_sectors_to_kill_list
 *
 * Used by both Legacy and Sysfs versions.
 **/
static int LD_add_sectors_to_kill_list( storage_object_t * ld,
					lsn_t lsn,
					sector_count_t count)
{
	int rc = EINVAL;

	LOG_PROC_ENTRY();

	if (lsn + count > ld->size) {
		LOG_ERROR("Sector offset plus count goes beyond the end of the disk.\n");
	} else {
		rc = EngFncs->add_sectors_to_kill_list(ld, lsn, count);
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * LD_commit_changes
 *
 * Disk manager doesn't do anything during commit. Just return success.
 **/
static int LD_commit_changes(storage_object_t * ld, commit_phase_t phase)
{
	LOG_PROC_ENTRY();
	LOG_PROC_EXIT_INT(0);
	return 0;
}

/**
 * LD_get_info
 *
 * Returns DISK specific information
 **/
static int LD_get_info(storage_object_t * object,
		       char * descriptor_name,
		       extended_info_array_t ** info)
{
	extended_info_array_t * Info;
	uint count = 0;
	int rc = EINVAL;

	LOG_PROC_ENTRY();

	if (!info) {
		goto out;
	}
	*info = NULL;
	
        if (descriptor_name) {
		goto out;
	}
	
	Info = EngFncs->engine_alloc(sizeof(extended_info_array_t) +
				     (LD_INFO_COUNT - 1) * sizeof(extended_info_t));
	if (!Info) {
		rc = ENOMEM;
		goto out;
	}
		
	Info->count = LD_INFO_COUNT;
		
	SET_STRING_FIELD(Info->info[LD_INFO_NAME_INDEX].name, "Name");
	SET_STRING_FIELD(Info->info[LD_INFO_NAME_INDEX].title, "Name");
	SET_STRING_FIELD(Info->info[LD_INFO_NAME_INDEX].desc, "EVMS name for the DISK storage object");
	Info->info[LD_INFO_NAME_INDEX].type = EVMS_Type_String;
	SET_STRING_FIELD(Info->info[LD_INFO_NAME_INDEX].value.s, object->name);

	SET_STRING_FIELD(Info->info[LD_INFO_SIZE_INDEX].name, "Size");
	SET_STRING_FIELD(Info->info[LD_INFO_SIZE_INDEX].title, "Size");
	SET_STRING_FIELD(Info->info[LD_INFO_SIZE_INDEX].desc, "Size of the disk in sectors");
	Info->info[LD_INFO_SIZE_INDEX].type = EVMS_Type_Unsigned_Int64;
	Info->info[LD_INFO_SIZE_INDEX].unit = EVMS_Unit_Sectors;
	Info->info[LD_INFO_SIZE_INDEX].value.ui64 = object->size;

	SET_STRING_FIELD(Info->info[LD_INFO_CYLINDERS_INDEX].name, "Cyl");
	SET_STRING_FIELD(Info->info[LD_INFO_CYLINDERS_INDEX].title, "Cylinders");
	SET_STRING_FIELD(Info->info[LD_INFO_CYLINDERS_INDEX].desc, "Drive geometry -- number of cylinders");
	Info->info[LD_INFO_CYLINDERS_INDEX].type = EVMS_Type_Unsigned_Int64;
	Info->info[LD_INFO_CYLINDERS_INDEX].value.ui64 = object->geometry.cylinders;

	SET_STRING_FIELD(Info->info[LD_INFO_HEADS_INDEX].name, "Heads");
	SET_STRING_FIELD(Info->info[LD_INFO_HEADS_INDEX].title, "Heads");
	SET_STRING_FIELD(Info->info[LD_INFO_HEADS_INDEX].desc, "Drive geometry -- number of heads");
	Info->info[LD_INFO_HEADS_INDEX].type = EVMS_Type_Unsigned_Int32;
	Info->info[LD_INFO_HEADS_INDEX].value.ui32 = object->geometry.heads;

	SET_STRING_FIELD(Info->info[LD_INFO_SECTORS_INDEX].name, "Sectors");
	SET_STRING_FIELD(Info->info[LD_INFO_SECTORS_INDEX].title, "Sectors");
	SET_STRING_FIELD(Info->info[LD_INFO_SECTORS_INDEX].desc, "Drive geometry -- sectors per track");
	Info->info[LD_INFO_SECTORS_INDEX].type = EVMS_Type_Unsigned_Int32;
	Info->info[LD_INFO_SECTORS_INDEX].value.ui32 = object->geometry.sectors_per_track;

	SET_STRING_FIELD(Info->info[LD_INFO_SECTOR_SIZE_INDEX].name, "SectorSize");
	SET_STRING_FIELD(Info->info[LD_INFO_SECTOR_SIZE_INDEX].title, "Sector Size");
	SET_STRING_FIELD(Info->info[LD_INFO_SECTOR_SIZE_INDEX].desc, "Number of bytes per sector");
	Info->info[LD_INFO_SECTOR_SIZE_INDEX].type = EVMS_Type_Unsigned_Int32;
	Info->info[LD_INFO_SECTOR_SIZE_INDEX].unit = EVMS_Unit_Bytes;
	Info->info[LD_INFO_SECTOR_SIZE_INDEX].value.ui32 = object->geometry.bytes_per_sector;

	SET_STRING_FIELD(Info->info[LD_INFO_BLOCK_SIZE_INDEX].name, "BlockSize");
	SET_STRING_FIELD(Info->info[LD_INFO_BLOCK_SIZE_INDEX].title, "Block Size");
	SET_STRING_FIELD(Info->info[LD_INFO_BLOCK_SIZE_INDEX].desc, "Number of bytes per block");
	Info->info[LD_INFO_BLOCK_SIZE_INDEX].type = EVMS_Type_Unsigned_Int64;
	Info->info[LD_INFO_BLOCK_SIZE_INDEX].unit = EVMS_Unit_Bytes;
	Info->info[LD_INFO_BLOCK_SIZE_INDEX].value.ui64 = object->geometry.block_size;

	SET_STRING_FIELD(Info->info[LD_INFO_BOOT_LIMIT_INDEX].name, "BootLimit");
	SET_STRING_FIELD(Info->info[LD_INFO_BOOT_LIMIT_INDEX].title, "Boot Cylinder Limit");
	SET_STRING_FIELD(Info->info[LD_INFO_BOOT_LIMIT_INDEX].desc, "LBA of the first sector above the boot cylinder limit for this drive");
	Info->info[LD_INFO_BOOT_LIMIT_INDEX].type = EVMS_Type_Unsigned_Int64;
	Info->info[LD_INFO_BOOT_LIMIT_INDEX].value.ui64 = object->geometry.boot_cylinder_limit;

	GetListSize(object->parent_objects, &count);

	SET_STRING_FIELD(Info->info[LD_INFO_SEGMENTS_INDEX].name, "Segments");
	SET_STRING_FIELD(Info->info[LD_INFO_SEGMENTS_INDEX].title, "Segments");
	SET_STRING_FIELD(Info->info[LD_INFO_SEGMENTS_INDEX].desc, "Number of segments discovered on the drive (metadata, data, freespace)");
	Info->info[LD_INFO_SEGMENTS_INDEX].type = EVMS_Type_Unsigned_Int32;
	Info->info[LD_INFO_SEGMENTS_INDEX].value.ui32 = count;

	*info = Info;
	rc = 0;

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * LD_get_plugin_info
 *
 * Returns plug-in specific information
 **/
static int LD_get_plugin_info(char * descriptor_name,
			      extended_info_array_t ** info)
{
	int rc = EINVAL;
	extended_info_array_t * Info;
	char version_string[64];
	char required_engine_api_version_string[64];
	char required_plugin_api_version_string[64];

	LOG_PROC_ENTRY();

	if (!info) {
		goto out;
	}
	*info = NULL;

	if (descriptor_name) {
		goto out;
	}

	Info = EngFncs->engine_alloc(sizeof(extended_info_array_t) +
				     6 * sizeof(extended_info_t));
	if (!Info) {
		rc = ENOMEM;
		goto out;
	}

	Info->count = 6;

	sprintf(version_string, "%d.%d.%d",
		MAJOR_VERSION, MINOR_VERSION, PATCH_LEVEL);

	sprintf(required_engine_api_version_string, "%d.%d.%d",
		LD_PluginRecord->required_engine_api_version.major,
		LD_PluginRecord->required_engine_api_version.minor,
		LD_PluginRecord->required_engine_api_version.patchlevel);

	sprintf(required_plugin_api_version_string, "%d.%d.%d",
		LD_PluginRecord->required_plugin_api_version.plugin.major,
		LD_PluginRecord->required_plugin_api_version.plugin.minor,
		LD_PluginRecord->required_plugin_api_version.plugin.patchlevel);

	SET_STRING_FIELD(Info->info[0].name, "Short Name");
	SET_STRING_FIELD(Info->info[0].title, "Short Name");
	SET_STRING_FIELD(Info->info[0].desc, "A short name given to this plugin");
	Info->info[0].type = EVMS_Type_String;
	SET_STRING_FIELD(Info->info[0].value.s, LD_PluginRecord->short_name);

	SET_STRING_FIELD(Info->info[1].name, "Long Name");
	SET_STRING_FIELD(Info->info[1].title, "Long Name");
	SET_STRING_FIELD(Info->info[1].desc, "A long name given to this plugin");
	Info->info[1].type = EVMS_Type_String;
	SET_STRING_FIELD(Info->info[1].value.s, LD_PluginRecord->long_name);

	SET_STRING_FIELD(Info->info[2].name, "Type");
	SET_STRING_FIELD(Info->info[2].title, "Plug-in Type");
	SET_STRING_FIELD(Info->info[2].desc, "There are various types of plugins, each responsible for some kind of storage object.");
	Info->info[2].type = EVMS_Type_String;
	SET_STRING_FIELD(Info->info[2].value.s, "Device Manager");

	SET_STRING_FIELD(Info->info[3].name, "Version");
	SET_STRING_FIELD(Info->info[3].title, "Plugin Version");
	SET_STRING_FIELD(Info->info[3].desc, "Version number of this plug-in");
	Info->info[3].type = EVMS_Type_String;
	SET_STRING_FIELD(Info->info[3].value.s, version_string);

	SET_STRING_FIELD(Info->info[4].name, "Required Engine Services Version");
	SET_STRING_FIELD(Info->info[4].title, "Required Engine Services Version");
	SET_STRING_FIELD(Info->info[4].desc, "Version of the Engine services that this plug-in requires. It will not run on older versions of the Engine services.");
	Info->info[4].type = EVMS_Type_String;
	SET_STRING_FIELD(Info->info[4].value.s, required_engine_api_version_string);

	SET_STRING_FIELD(Info->info[5].name, "Required Plug-in API Version");
	SET_STRING_FIELD(Info->info[5].title, "Required Plug-in API Version");
	SET_STRING_FIELD(Info->info[5].desc, "Version of the Engine plug-in API that this plug-in requires. It will not run on older versions of the Engine plug-in API.");
	Info->info[5].type = EVMS_Type_String;
	SET_STRING_FIELD(Info->info[5].value.s, required_plugin_api_version_string);

	*info = Info;
	rc = 0;

out:
	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/**
 * LD_direct_plugin_communication
 *
 * Device-manager-specific commands for starting and stopping the cache, and
 * opening and closing a disk.
 **/
static int LD_direct_plugin_communication(void * object,
					  BOOLEAN target_kernel_plugin,
					  void * arg)
{
	int rc = 0, fd = -1;
	storage_object_t * ld = object;
	ldm_dpc_t * dcp = arg;

	LOG_PROC_ENTRY();

	switch (dcp->function) {

	case START_CACHE:
		initialize_cache();
		break;

	case STOP_CACHE:
		destroy_cache();
		break;

       	case LDM_CLOSE_DISK:
		fd = *(int *)ld->private_data;
		if (fd >= 0) {			
			rc = EngFncs->close_object(ld, fd);
			if (!rc) {
				*(int *)ld->private_data = -1;
			}
		}
		break;

	case LDM_OPEN_DISK:
		fd = *(int *)ld->private_data;
		if (fd < 0) {	
			fd = EngFncs->open_object(ld, O_DIRECT|O_RDWR|O_SYNC);
			if (fd >= 0) {	
				*(int *)ld->private_data = fd;
				rc = 0;
			} else {
				rc = errno;
			}
		}
		break;

	default:
		LOG_ERROR("%d is not a valid function code.\n", dcp->function);
		rc = EINVAL;
	}

	LOG_PROC_EXIT_INT(rc);
	return rc;
}

/********** Legacy Version **********/
 
static plugin_functions_t ft_legacy = {
	.setup_evms_plugin		= LD_setup_legacy,
	.cleanup_evms_plugin		= LD_cleanup,
	.discover			= LD_discover_legacy,
	.add_sectors_to_kill_list	= LD_add_sectors_to_kill_list,
	.commit_changes			= LD_commit_changes,
	.get_info			= LD_get_info,
	.get_plugin_info		= LD_get_plugin_info,
	.read				= LD_read,
	.write				= LD_write,
	.direct_plugin_communication	= LD_direct_plugin_communication
};

plugin_record_t LD_PluginRecord_legacy = {
	.id = SetPluginID(EVMS_OEM_IBM, EVMS_DEVICE_MANAGER, 3),
	.version = {
		.major		= MAJOR_VERSION,
		.minor		= MINOR_VERSION,
		.patchlevel	= PATCH_LEVEL
	},
	.required_engine_api_version = {
		.major		= 9,
		.minor		= 0,
		.patchlevel	= 0
	},
	.required_plugin_api_version = {
		.plugin = {
			.major		= 10,
			.minor		= 0,
			.patchlevel	= 0
		}
	},
	.short_name = "LocalDskMgr",
	.long_name = "Local Disk Manager",
	.oem_name = "IBM",
	.functions = {
		.plugin = &ft_legacy
	},
	.container_functions = NULL
};

/********** Sysfs Version **********/

static plugin_functions_t ft_sysfs = {
	.setup_evms_plugin		= LD_setup_sysfs,
	.cleanup_evms_plugin		= LD_cleanup,
	.discover			= LD_discover_sysfs,
	.add_sectors_to_kill_list	= LD_add_sectors_to_kill_list,
	.commit_changes			= LD_commit_changes,
	.get_info			= LD_get_info,
	.get_plugin_info		= LD_get_plugin_info,
	.read				= LD_read,
	.write				= LD_write,
	.direct_plugin_communication	= LD_direct_plugin_communication
};

plugin_record_t LD_PluginRecord_sysfs = {
	.id = SetPluginID(EVMS_OEM_IBM, EVMS_DEVICE_MANAGER, 2),
	.version = {
		.major		= MAJOR_VERSION,
		.minor		= MINOR_VERSION,
		.patchlevel	= PATCH_LEVEL
	},
	.required_engine_api_version = {
		.major		= 9,
		.minor		= 0,
		.patchlevel	= 0
	},
	.required_plugin_api_version = {
		.plugin = {
			.major		= 10,
			.minor		= 0,
			.patchlevel	= 0
		}
	},
	.short_name = "LocalDskMgr",
	.long_name = "Local Disk Manager",
	.oem_name = "IBM",
	.functions = {
		.plugin = &ft_sysfs
	},
	.container_functions = NULL
};

plugin_record_t * plugin_legacy = &LD_PluginRecord_legacy;
plugin_record_t * plugin_sysfs = &LD_PluginRecord_sysfs;
plugin_record_t * LD_PluginRecord = NULL;
plugin_record_t * evms_plugin_records[] = { &LD_PluginRecord_sysfs,
					    &LD_PluginRecord_legacy,
					    NULL };

