/*
 *   Copyright (c) International Business Machines  Corp., 2001
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or 
 *   (at your option) any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software 
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: lvmregmgr
 * File: lvm_move.c
 *
 * Description: This file contains all functions related to moving extents.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <plugin.h>
#include "lvmregmgr.h"

static int disable_move_message = FALSE;

/**
 * lvm_move_extent_setup
 * @le:		Logical extent to move.
 * @dst_pe:	Physical extent to move the LE to.
 *
 * Setup the specified LE to be moved from its current PE to the new PE. Make
 * sure the destination PE is available before calling this function.
 **/
static int lvm_move_extent_setup(lvm_logical_extent_t * le,
				 lvm_physical_extent_t * dst_pe)
{
	lvm_logical_volume_t * volume = le->volume;
	lvm_volume_group_t * group = volume->group;

	LOG_ENTRY;

	le->new_pe = dst_pe;
	dst_pe->new_le = le;
	dst_pe->pv->move_extents++;
	group->move_extents++;
	volume->flags |= LVM_LV_FLAG_MOVE_PENDING;

	lvm_append_region_to_segment(volume->region, dst_pe->pv->segment);

	RETURN(0);
}

/**
 * lvm_commit_start_copy_progress
 *
 * Set-up the copy-progress indicator.
 **/
static int lvm_commit_start_copy_progress(copy_job_t * copy_job,
					  progress_t * progress)
{
	int title_size = (EVMS_NAME_SIZE+5)*3 + 30;
	lvm_logical_extent_t * le = copy_job->context;

	LOG_ENTRY;

	progress->id = 0;
	progress->title = lvm_engine->engine_alloc(title_size);
	snprintf(progress->title, title_size,
		 "LVM: Copying LE %s:%d from %s:%d to %s:%d", 
		 le->volume->region->name, le->number,
		 le->pe->pv->segment->name, le->pe->number,
		 le->new_pe->pv->segment->name, le->new_pe->number);
	progress->description = NULL;
	progress->type = DISPLAY_PERCENT;
	progress->count = 0;
	progress->total_count = copy_job->src.len / copy_job->chunk_size + 1;
	progress->remaining_seconds = 0;
	progress->plugin_private_data = le;

	copy_job->context = progress;

	lvm_engine->progress(progress);

	RETURN(0);
}

/**
 * lvm_commit_update_copy_progress
 *
 * Status function that will be called periodically during the copying of
 * an extent.
 **/
static void lvm_commit_update_copy_progress(u_int64_t chunks_copied,
					    void * context)
{
	progress_t * progress = context;
	progress->count++;
	lvm_engine->progress(progress);
}

/**
 * lvm_commit_end_copy_progress
 *
 * Tear-down the copy-progress indicator.
 **/
static int lvm_commit_end_copy_progress(progress_t * progress)
{
	LOG_ENTRY;

	progress->count = progress->total_count;
	lvm_engine->progress(progress);
	lvm_engine->engine_free(progress->title);

	RETURN(0);
}

/**
 * lvm_commit_copy_extent
 *
 * Copy the data from the source PE to the destination PE.
 *
 * Offline implementation:
 * Setup an engine copy job and submit to the engine's copy-service.
 *
 * Online implementation:
 * To be written...
 **/
static int lvm_commit_copy_extent(lvm_physical_extent_t * src_pe,
				  lvm_physical_extent_t * dst_pe)
{
	copy_job_t copy_job;
	progress_t copy_progress;
	u_int32_t pe_size = src_pe->pv->pv->pe_size;
	int rc;

	LOG_ENTRY;

	copy_job.src.obj = src_pe->pv->segment;
	copy_job.src.start = src_pe->sector;
	copy_job.src.len = pe_size;
	copy_job.trg.obj = dst_pe->pv->segment;
	copy_job.trg.start = dst_pe->sector;
	copy_job.trg.len = pe_size;
	copy_job.chunk_size = min(pe_size, LVM_COPY_CHUNK_SIZE_DEFAULT);
	copy_job.context = src_pe->le;

	lvm_commit_start_copy_progress(&copy_job, &copy_progress);
	rc = lvm_engine->copy(&copy_job, lvm_commit_update_copy_progress);
	lvm_commit_end_copy_progress(&copy_progress);

	RETURN(rc);
}

/**
 * lvm_commit_move_extent
 *
 * Move the specified LE from its current PE to its new PE. Copy the data
 * from the old PE to the new PE, update the mappings for both PEs, write
 * out the new metadata, and reactivate the volume with the new mapping.
 **/
static int lvm_commit_move_extent(lvm_logical_extent_t * le)
{
	lvm_physical_extent_t * src_pe = le->pe;
	lvm_physical_extent_t * dst_pe = le->new_pe;
	int rc = 0;

	LOG_ENTRY;

	if (!dst_pe) {
		RETURN(0);
	}

	/* Copy the data from the source to the destination. */
	if (src_pe) {
		rc = lvm_commit_copy_extent(src_pe, dst_pe);
		if (rc) {
			RETURN(rc);
		}
	}

	/* Update the source's metadata. */
	if (src_pe) {
		src_pe->pe.lv_num = 0;
		src_pe->pe.le_num = 0;
		src_pe->pv->pv->pe_allocated--;
		if (!lvm_volume_is_on_pv(le->volume, src_pe->pv)) {
			src_pe->pv->pv->lv_cur--;
			lvm_remove_region_from_segment(le->volume->region,
						       src_pe->pv->segment);
		}
		src_pe->le = NULL;
	}

	/* Update the destination's metadata. */
	if (!lvm_volume_is_on_pv(le->volume, dst_pe->pv)) {
		dst_pe->pv->pv->lv_cur++;
	}
	dst_pe->pe.lv_num = le->volume->number;
	dst_pe->pe.le_num = le->number;
	dst_pe->pv->pv->pe_allocated++;
	dst_pe->new_le = NULL;
	dst_pe->le = le;
	dst_pe->pv->move_extents--;
	dst_pe->pv->group->move_extents--;

	/* Update volume's map. */
	le->pe = dst_pe;
	le->new_pe = NULL;

	/* Write the metadata to disk. */
	if (src_pe) {
		lvm_write_pv(src_pe->pv);
		lvm_write_pe_map(src_pe->pv);
	}
	lvm_write_pv(dst_pe->pv);
	lvm_write_pe_map(dst_pe->pv);

	/* Re-activate the volume with the new mapping. */
	rc = lvm_plugin->functions.plugin->activate(le->volume->region);

	RETURN(rc);
}

/**
 * lvm_volume_is_busy
 *
 * Check if the specified volume is open/in-use/mounted. If so, ask the user
 * if they'd like to unmount and retry.
 *
 * Return: TRUE: skip processing this volume.
 *         FALSE: proceed with this volume.
 **/
static int lvm_volume_is_busy(lvm_logical_volume_t * volume)
{
	char * choices[] = {"Skip", "Retry", NULL};
	char * mount_point;
	int mounted, answer;
	int rc = FALSE;

	LOG_ENTRY;

	if (!volume->region->volume) {
		/* If the region does not have an EVMS volume, assume
		 * the user knows what he is doing.
		 */
		goto out;
	}

	/* This loop will not exit unless the volume is unmounted, or the
	 * user decides to skip this volume.
	 */
	while ((mounted = lvm_engine->is_mounted(volume->region->volume->name,
						 &mount_point))) {
		answer = 0;
		lvm_engine->user_message(lvm_plugin, &answer, choices,
					 "Region \"%s\" has extents scheduled "
					 "to be moved. However, this region is "
					 "part of volume \"%s\", which is "
					 "mounted at %s. Please unmount the "
					 "volume and choose \"%s\" to continue "
					 "with the move, or choose \"%s\" to "
					 "skip the move at this time (the move "
					 "will be attempted again the next time "
					 "changes are saved).",
					 volume->region->name,
					 volume->region->volume->name,
					 mount_point, choices[1], choices[0]);
		if (answer == 0) {
			break;
		}
	}

	rc = (mounted && answer == 0);

out:
	RETURN(rc);
}

/**
 * lvm_commit_move_extents
 *
 * Move all extents that are waiting to be moved in this container.
 **/
int lvm_commit_move_extents(lvm_volume_group_t * group)
{
	lvm_logical_volume_t * volume;
	int i, j, rc = 0;

	LOG_ENTRY;

	for (i = 1; i < MAX_LV; i++) {
		volume = group->volume_list[i];
		if (volume && (volume->flags & LVM_LV_FLAG_MOVE_PENDING)) {

			rc = lvm_volume_is_busy(volume);
			if (rc) {
				continue;
			}

			for (j = 0; j < volume->lv->lv_allocated_le; j++) {
				if (volume->le_map[j].new_pe) {
					rc |= lvm_commit_move_extent(&volume->le_map[j]);
				}
			}

			if (!rc) {
				volume->flags &= ~LVM_LV_FLAG_MOVE_PENDING;
			}
		}
	}

	/* Need to rebuild the freespace, since the number and
	 * location of unused extents has probably changed.
	 */
	rc = lvm_update_freespace_volume(group);
	if (rc) {
		LOG_SERIOUS("Error updating freespace for container %s\n",
			    group->container->name);
	}

	RETURN(rc);
}


/********** Move-Extent Task Functions **********/


/**
 * lvm_region_move_extent_init_option_descriptor
 *
 * Initialize the option descriptor for a MOVE_EXTENT task.
 * Option 0: Logical Extent
 * Option 1: Physical Volume
 * Option 2: Physical Extent
 **/
int lvm_region_move_extent_init_option_descriptor(task_context_t * context)
{
	option_desc_array_t * od = context->option_descriptors;
	lvm_logical_volume_t * volume = context->object->private_data;
	lvm_volume_group_t * group = volume->group;
	lvm_physical_volume_t * pv_entry;
	u_int32_t i, index, count = 0;

	LOG_ENTRY;

	/* Option 0: logical extent number.
	 * Generate a list of all the extents in this
	 * volume that are not already being moved.
	 */
	index = LVM_OPTION_MOVE_EXTENT_LE_INDEX;
	SET_STRING(od->option[index].name, LVM_OPTION_MOVE_EXTENT_LE_STR);
	SET_STRING(od->option[index].title, "Logical Extent");
	SET_STRING(od->option[index].tip, "Logical extent to move");
	od->option[index].type = EVMS_Type_Unsigned_Int32;
	od->option[index].flags = EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;
	od->option[index].constraint_type = EVMS_Collection_List;
	od->option[index].constraint.list = lvm_engine->engine_alloc(sizeof(value_list_t) +
								     sizeof(value_t) *
								     volume->lv->lv_allocated_le);
	for (i = 0; i < volume->lv->lv_allocated_le; i++) {
		if (!lvm_le_is_scheduled_for_move(&volume->le_map[i])) {
			od->option[index].constraint.list->value[count].ui32 = i;
			count++;
		}
	}
	od->option[index].constraint.list->count = count;

	/* Option 1: new physical volume.
	 * Generate a list of all PVs that have unused extents.
	 */
	count = 0;
	index = LVM_OPTION_MOVE_EXTENT_PV_INDEX;
	SET_STRING(od->option[index].name, LVM_OPTION_MOVE_EXTENT_PV_STR);
	SET_STRING(od->option[index].title, "Physical Volume");
	SET_STRING(od->option[index].tip, "Physical volume to move this logical extent to.");
	od->option[index].type = EVMS_Type_String;
	od->option[index].max_len = EVMS_NAME_SIZE;
	od->option[index].flags = EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;
	od->option[index].constraint_type = EVMS_Collection_List;
	od->option[index].constraint.list = lvm_engine->engine_alloc(sizeof(value_list_t) +
								     sizeof(value_t) * MAX_PV);
	for (i = 0; i < MAX_PV; i++) {
		pv_entry = group->pv_list[i];
		if (pv_entry && lvm_pv_has_available_extents(pv_entry)) {
			SET_STRING(od->option[index].constraint.list->value[count].s,
				   pv_entry->segment->name);
			count++;
		}
	}
	od->option[index].constraint.list->count = count;
	od->option[index].value.s = lvm_engine->engine_alloc(EVMS_NAME_SIZE+1);

	/* Option 2: new physical extent.
	 * Set up a list option, but can't fill it in yet. Have to wait
	 * until the user chooses which PV to move this extent to.
	 */
	index = LVM_OPTION_MOVE_EXTENT_PE_INDEX;
	SET_STRING(od->option[index].name, LVM_OPTION_MOVE_EXTENT_PE_STR);
	SET_STRING(od->option[index].title, "Physical Extent");
	SET_STRING(od->option[index].tip, "Physical extent to move this logical extent to.");
	od->option[index].type = EVMS_Type_Unsigned_Int32;
	od->option[index].flags = EVMS_OPTION_FLAGS_NO_INITIAL_VALUE |
				  EVMS_OPTION_FLAGS_INACTIVE;
	od->option[index].constraint_type = EVMS_Collection_List;

	od->count = index + 1;
	RETURN(0);
}

/**
 * lvm_region_move_extent_set_option
 *
 * Verify and set an option during a MOVE_EXTENT task.
 **/
int lvm_region_move_extent_set_option(task_context_t * context,
				      u_int32_t index,
				      value_t * value,
				      task_effect_t * effect)
{
	option_desc_array_t * od = context->option_descriptors;
	lvm_logical_volume_t * volume = context->object->private_data;
	lvm_physical_volume_t * pv_entry;
	u_int32_t i, count = 0;
	int rc = EINVAL;

	LOG_ENTRY;
	LOG_EXTRA("Setting option %d\n", index);

	*effect = 0;

	switch (index) {

	case LVM_OPTION_MOVE_EXTENT_LE_INDEX:
		/* Make sure the LE is in range. */
		if (lvm_le_is_valid(volume, value->ui32) &&
		    !lvm_le_is_scheduled_for_move(&volume->le_map[value->ui32])) {
			od->option[index].value.ui32 = value->ui32;
			rc = 0;
		}
		break;

	case LVM_OPTION_MOVE_EXTENT_PV_INDEX:
		/* Make sure a valid PV was selected, and
		 * that the PV has unused PEs available.
		 */
		pv_entry = lvm_get_pv_for_name(value->s, volume->group);
		if (pv_entry && lvm_pv_has_available_extents(pv_entry)) {
			strncpy(od->option[index].value.s, value->s, EVMS_NAME_SIZE);

			/* Set up the list of possible PEs on this PV. */
			index = LVM_OPTION_MOVE_EXTENT_PE_INDEX;
			lvm_engine->engine_free(od->option[index].constraint.list);
			od->option[index].constraint.list = lvm_engine->engine_alloc(sizeof(value_list_t) +
										     sizeof(value_t) *
										     pv_entry->pv->pe_total);
			for (i = 0; i < pv_entry->pv->pe_total; i++) {
				if (lvm_pe_is_available(&pv_entry->pe_map[i])) {
					od->option[index].constraint.list->value[count].ui32 = i;
					count++;
				}
			}
			od->option[index].constraint.list->count = count;
			od->option[index].flags &= ~EVMS_OPTION_FLAGS_INACTIVE;
			od->option[index].flags |= EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;

			*effect |= EVMS_Effect_Reload_Options;
			rc = 0;
		}
		break;

	case LVM_OPTION_MOVE_EXTENT_PE_INDEX:
		/* Get the selected PV from the option descriptor,
		 * and make sure the selected PE is available.
		 */
		pv_entry = lvm_get_pv_for_name(od->option[LVM_OPTION_MOVE_EXTENT_PV_INDEX].value.s,
					       volume->group);
		if (pv_entry &&
		    lvm_pe_is_valid(pv_entry, value->ui32) &&
		    lvm_pe_is_available(&pv_entry->pe_map[value->ui32])) {
			od->option[index].value.ui32 = value->ui32;
			rc = 0;
		}
		break;

	default:
		break;
	}

	RETURN(rc);
}

/**
 * lvm_region_move_extent_parse_options
 **/
static int lvm_region_move_extent_parse_options(option_array_t * options,
						lvm_logical_volume_t * volume,
						lvm_physical_volume_t ** pv_entry,
						u_int32_t * le,
						u_int32_t * pe)
{
	int i;

	LOG_ENTRY;

	for (i = 0; i < options->count; i++) {
		/* If only the name-based index is specified, get the number. */
		if (!options->option[i].is_number_based) {
			if (!strcmp(options->option[i].name, LVM_OPTION_MOVE_EXTENT_LE_STR)) {
				options->option[i].number = LVM_OPTION_MOVE_EXTENT_LE_INDEX;
			} else if (!strcmp(options->option[i].name, LVM_OPTION_MOVE_EXTENT_PV_STR)) {
				options->option[i].number = LVM_OPTION_MOVE_EXTENT_PV_INDEX;
			} else if (!strcmp(options->option[i].name, LVM_OPTION_MOVE_EXTENT_PE_STR)) {
				options->option[i].number = LVM_OPTION_MOVE_EXTENT_PE_INDEX;
			} else {
				continue;
			}
		}

		LOG_EXTRA("Parsing option %d\n", options->option[i].number);

		/* Use the number-based index to record the option. */
		switch (options->option[i].number) {
		case LVM_OPTION_MOVE_EXTENT_LE_INDEX:
			*le = options->option[i].value.ui32;
			break;
		case LVM_OPTION_MOVE_EXTENT_PV_INDEX:
			*pv_entry = lvm_get_pv_for_name(options->option[i].value.s,
							volume->group);
			break;
		case LVM_OPTION_MOVE_EXTENT_PE_INDEX:
			*pe = options->option[i].value.ui32;
			break;
		default:
			break;
		}
	}

	RETURN(0);
}

/**
 * lvm_region_move_extent_verify_options
 *
 * Verify that the specified options are all valid for moving an extent.
 **/
int lvm_region_move_extent_verify_options(lvm_logical_volume_t * volume,
					  lvm_physical_volume_t * pv_entry,
					  u_int32_t le,
					  u_int32_t pe)
{
	int rc = EINVAL;

	LOG_ENTRY;

	if (!lvm_le_is_valid(volume, le)) {
		LOG_ERROR("LE %d is out of range for region %s.\n",
			  le, volume->region->name);
	} else if (lvm_le_is_scheduled_for_move(&volume->le_map[le])) {
		LOG_ERROR("LE %d on region %s is already scheduled for move.\n",
			  le, volume->region->name);
	} else if (!pv_entry) {
		LOG_ERROR("Invalid PV specified.\n");
	} else if (!lvm_pe_is_valid(pv_entry, pe)) {
		LOG_ERROR("PE %d is out of range for PV %s.\n",
			  pe, pv_entry->segment->name);
	} else if (!lvm_pe_is_available(&pv_entry->pe_map[pe])) {
		LOG_ERROR("PE %d on PV %s is not available for move.\n",
			  pe, pv_entry->segment->name);
	} else {
		rc = 0;
	}

	RETURN(rc);
}

/**
 * lvm_region_move_extent
 **/
int lvm_region_move_extent(lvm_logical_volume_t * volume,
			   option_array_t * options)
{
	lvm_physical_volume_t * pv_entry = NULL;
	u_int32_t le_num = 0, pe_num = 0;
	char * choices[] = {"Continue",
			    "Continue All",
			    "Don't Continue",
			    NULL};
	int rc, answer = 0;

	LOG_ENTRY;

	/* Parse and verify the options for move. */
	lvm_region_move_extent_parse_options(options, volume,
					     &pv_entry, &le_num, &pe_num);
	rc = lvm_region_move_extent_verify_options(volume, pv_entry,
						   le_num, pe_num);
	if (rc) {
		LOG_ERROR("Error verifying move-extent options.\n");
		RETURN(rc);
	}

	if (!disable_move_message) {
		lvm_engine->user_message(lvm_plugin, &answer, choices,
					 "You have chosen to move LE %s:%d from "
					 "PE %s:%d to PE %s:%d. This move cannot "
					 "be performed safely while region %s is "
					 "in use. Please be sure %s is unmounted "
					 "and not in use before saving these "
					 "changes.",
					 volume->region->name, le_num,
					 volume->le_map[le_num].pe ? volume->le_map[le_num].pe->pv->segment->name : "missing-pv",
					 volume->le_map[le_num].pe ? volume->le_map[le_num].pe->number : 0, 
					 pv_entry->segment->name, pe_num,
					 volume->region->name,
					 volume->region->name);
		if (answer == 1) {
			disable_move_message = TRUE;
		} else if (answer == 2) {
			RETURN(E_CANCELED);
		}
	}

	/* Prepare the extents to be moved. */
	lvm_move_extent_setup(&volume->le_map[le_num], &pv_entry->pe_map[pe_num]);

	/* Rebuild the freespace. */
	rc = lvm_update_freespace_volume(volume->group);
	if (rc) {
		LOG_ERROR("Error updating freespace for container %s\n",
			  volume->group->container->name);
	}

	volume->group->container->flags |= SCFLAG_DIRTY;
	lvm_engine->set_changes_pending();

	RETURN(rc);
}

