/*
 *
 *   Copyright (c) International Business Machines  Corp., 2002, 2003
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or 
 *   (at your option) any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software 
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */ 

#include <string.h>
#include <glib.h>
#include <frontend.h>
#include <ncurses.h>
#include <panel.h>

#include "common.h"
#include "dlist.h"
#include "menu.h"
#include "window.h"
#include "dialog.h"
#include "callbacks.h"

static int queue_message = FALSE;
static GList *message_queue;

/**
 *	queue_informational_message - places an informational message in queue
 *	@message: the information message to duplicate and queue
 *
 *	This routine allocates a queue element, duplicates the message and adds
 *	the alert message queue element to the queue.
 */
void queue_informational_message(char *message)
{
	if (message != NULL)
		message_queue = g_list_append(message_queue, g_strdup(message));
}

/**
 *	delete_queued_message - deletes a particular message from the queue
 *	@message: the address of the message to delete
 *
 *	This routine removes a message in the alert message queue and
 *	frees the memory for the message and the message queue element.
 */
void delete_queued_message(char *message)
{
	if (message != NULL) {
		message_queue = g_list_remove(message_queue, message);
		g_free(message);
	}
}

/**
 *	enable_message_queuing - allows queuing of information alerts
 *	@void
 *
 *	This routine simply sets a flag that indicates to the show_user_message()
 *	routine to queue information only alert messages rather than immediately
 *	display them.
 **/
void enable_message_queuing(void)
{
	queue_message = TRUE;
}

/**
 *	disable_message_queuing - stop queuing information only alert messages
 *	@void
 *
 *	This routine simply turns the flags off that indicates that alert messages
 *	should be queued rather than displayed immediately.
 **/
void disable_message_queuing(void)
{
	queue_message = FALSE;
}

/**
 *	report_progress - callback to update statusbar with progress by engine
 *	@progress: the address of the structure containing progress info
 *
 *	This routine is called by the engine to provide operation progress info
 *	typically originated by a plugin.
 **/
int report_progress(progress_t *progress)
{
	if (progress->title) {
		print_statusbar_text(progress->title);

		if (progress->type == DISPLAY_PERCENT) {
			int startx;
			char percent_complete[6];

			if (progress->total_count == 0) {
				strcpy(percent_complete, "0%");
			} else {
				float percentage;

				percentage = ((float)progress->count / (float)progress->total_count) * 100;
				g_snprintf(percent_complete, sizeof(percent_complete), " %.0f%%",  percentage);
			}
			startx = getmaxx(stdscr) - strlen(percent_complete) - 2;
			mvwaddstr(get_statusbar_win(), 0, startx, percent_complete);
		}
	}
 	return 0;
}

/**
 *	on_alert_choice_deleted - callback for when alert choice deleted
 *	@item: the menu item being deleted
 *
 *	This routine simply frees the alert_choice_info struct
 *	associated with a alert choice menu item.
 */
void on_alert_choice_deleted(struct menu_item *item)
{
	g_free(item->user_data);
}

/**
 *	alert_choice_activated - callback for when alert choice is activated
 *	@item: the menu item that was activated
 *
 *	This routine simply updates the alert's return selection
 *	to the index of the activated choice.
 */
int alert_choice_activated(struct menu_item *item)
{
	struct alert_choice_info *info = item->user_data;

	*(info->return_selection) = info->our_index;

	return 0;
}

/**
 *	append_alert_choice - append a choice to the menu of choices for an alert window
 *	@menu: the horizontal menu
 *	@choice: the string containing the choice
 *	@return_selection: where to place index of choice selection. initialized with default choice.
 *	@our_index: the index value for our choice
 *
 *	This routine appends one alert message choice to the horizontal menu of choices.
 *	It sets up the callbacks for the alert choice and also sets focus on this choice
 *	if it is the default choice.
 **/
void append_alert_choice(struct horizontal_menu *menu, char *choice, int *return_selection, int our_index)
{
	char *choice_button;
	struct menu_item *item;
	struct alert_choice_info *info;
	
	info = g_new(struct alert_choice_info, 1);
	info->our_index = our_index;
	info->return_selection = return_selection;

	choice_button = g_strdup_printf("[%s]", choice);

	item = pack_menu_item_at_end(menu, choice_button, 0,
				(menuitem_activate_cb)alert_choice_activated, 
				info);

	set_menu_item_delete_cb(item, (menuitem_delete_cb)on_alert_choice_deleted);

	/*
	 * If the default return selection matches this one then make
	 * it the focus item.
	 */
	if (*return_selection == our_index) {
		set_horizontal_menu_focus(menu, item);		
	}

	g_free(choice_button);
}

/**
 *	show_alert_prompt - prompt a user with a set of choices a plug-in has supplied
 *	@message: the alert message to display
 *	@choice_selection: the index of the choice selection
 *	@choices: array of choices
 *
 *	This routine creates a popup window with an alert message a set of choices
 *	for the user. The popup window is removed when the user selects a choice.
 */
void show_alert_prompt(char *message, int *choice_selection, char **choices)
{
	int key = 0, i;
	PANEL *panel;
	WINDOW *win;
	struct horizontal_menu *menu;

	panel = create_centered_popup_window(getmaxy(stdscr) - 2, 
					getmaxx(stdscr) - 6, WHITE_BKGD);
	win = panel_window(panel);

	mvwprintw(win, 0, 0, message);

	menu = create_horizontal_menu(win, getmaxx(win) - 4, getmaxy(win) - 2, 1, 1);
	
	for (i = 0; choices[i] != NULL; i++) {
		append_alert_choice(menu, choices[i], choice_selection, i);
	}

	draw_horizontal_menu(menu);
	show_popup_window(panel);

	while (key != KEY_ENTER) {
		key = panel_getch(panel);
		process_horizontal_menu_events(menu, &key);
	}

	delete_horizontal_menu(menu);	
	delete_popup_window(panel);
}

/**
 *	show_user_message - callback to allow engine or engine plugin to alert user
 *	@message: the alert message to display
 *	@choice_selection: the index of the choice selection if choices given
 *	@choices: array of choices for an interactive alert
 *
 *	This routine is called by the engine when it or a plugin needs to display an
 *	alert to a user or when the user needs to be asked for a response from a set
 *	of choices. During engine open, we are asked to queue messages. This allows
 *	the user to receive one dialog to allow them to see the informational messages
 *	that were queued once the main window is up.
 */
int show_user_message(char *message, int *choice_selection, char **choices)
{
	if (choices == NULL) {
		if (queue_message == TRUE) {
			queue_informational_message(message);
		} else {
			show_message_dialog(NULL, message);
		}
	} else {
		show_alert_prompt(message, choice_selection, choices);
	}
	return 0;
}

/**
 *	display_informational_message - display one message
 *	@message: the informational message
 *	@not_used: extra data not used at this time
 */
void display_informational_message(gchar *message, gpointer not_used)
{
	show_message_dialog(NULL, message);
	g_free(message);
}

/**
 *	display_queued_messages - allows the user to view messages we previously queued
 *	@void
 *
 *	This routine allows the user to view information alert messages we received
 *	and queued during some engine operation.
 **/
void display_queued_messages(void)
{
	if (g_list_length(message_queue) > 0) {
		delete_all_elements(message_queue, (GFunc)display_informational_message, NULL);
		message_queue = NULL;
	}
}
